<?php
/**
 * Change the text file to Songbook (ChordPro) format
 */

class SongBook{
  var $cli;
  var $test = False;
  var $fname = "./songbook_test.pro";
  var $rows = array();
  var $chg   = array();
  var $h;
  var $hs = [
  't'        => '/^\{(t:|title:).*\}/i',	
	'st'       => '/^\{(st:|subtitle:).*\}/i',
  'artist'   => '/^\{artist:.*\}/i',
  'key'      => '/^\{key:.*\}/i',
	'time'     => '/^\{time:.*\}/i',
	'duration' => '/^\{duration:.*\}/i',
	'tempo'    => '/^\{tempo:.*\}/i',
  'metronome'=> '/^\{metronome:.*\}/i',
  'tag'      => '/^\{tag:.*\}/i',
  'musicpath'=> '/^\{musicpath:.*\}/i',
  'zoom'     => '/^\{zoom:.*\}/i',
  ];
  var $headers =  [
      't'        => '{t:}',
      'st'       => '{st:}',
      'artist'   => '{artist:}',
      'key'      => '{key:}',
      'time'     => '{time:0:01}',
      'duration' => '{duration:0:01}',
      'tempo'    => '{tempo:120}',
      'metronome'=> '{metronome: 120}',
      'tag'      => '{tag: }',
      'musicpath'=> '{musicpath:}',
      'zoom'     => '{zoom:}',
      'artist'   => '{artist:}',
  ];
  var $LastRow = false;
  var $InsertChordNextLine = false;   //Insert the [chord] in the next correct place of next line 
  var $size0     = 0;
  var $size1     = 0;
  var $t         = "";
  var $st        = "";
  var $key       = "";
  var $time      = "";
  var $musicpath = "";
  var $tag       = "";
  var $duration  = "";
  var $tags_cl   = array();
  var $sot = false;
  var $pre = "<<<";
  var $post = ">>>";
  var $max = 0;
  var $instrument = "Guitar";

  var $Tones= array("C","C#","Db","D","D#","Eb","E","F","F#","Gb","G","G#","Ab","A","A#","Bb","B","B#","H");
  var $chords   = array(
      "maj7b5", "maj7", "maj", "m5", "m6", "m7", "m9", "m", "dim", "add", "7sus4", "msus4","madd9",
      "sus2" ,"sus4","sus", "9sus4", "7-9","7sus2","7sus","mM7","\+","aug","4" , "5" , "6" , "7" ,"9-5","9", "0",""
  );
  
  //Refrain, verse, stb. modifying
  var $words= array(
      "refr\.", "Refrén", "Refren","Refrain 3", "Refrain 2", "Refrain 1","refrain", "R1", "R2", "Melodie - Intro",
      "Intro 2", "Intro 1", "Intro", "Riff", "Coda",
      "Versszak","Verse 1", "Verse 2", "Verse 3", "Verse 4", "Verse 5", "Verse 6", "Verse 7", "Verse 8","Verse", 
      "Bridge", "Post-Chorus", "Pre-Chorus 1", "Pre-Chorus", "Chorus 2", "chorus 1", "Chorus 3", "Chorus\/Outro", "Chorus", 
      "Guitar solo", "Solo", "Git�r sz�l�", "Sz�l�","Turnaround",
      "Link and Outro", "Outro","outro", "Instrumental", "Interlude 2", "Interlude 1", "Interlude", "Mid-section",
      "Middle 8", "Middle 7", "Middle 6", "Middle 5", "Middle 4", "Middle 3", "Middle 2", "Middle 1", "Middle"
  );
  
  function __construct($fname = "./songbook_test.pro"){    
    $this->cli = isset($_SERVER["argv"])?true:False;        
    if(!empty($fname)){
      $this->fname =$fname; 
    }
    if(isset( $_SESSION )) {
      $_SESSION["fname"] = $this->fname;
    }
    
    $p = pathinfo($this->fname);    
    if(!isset($p["extension"])){
      $this->fname .= ".pro";
    }
    if(isset($p["extension"]) && strtolower($p["extension"]) !="pro"){
      $this->fname .= ".pro";
    }
    if(empty($this->fname)) 
      $this->Help("There is no filename!");
    //
    $this->ReadTags();
    $p = getcwd()."/".$this->fname;
    if(file_exists($p)){
      $this->Load();
    }else{
      $this->Help("\nThere is no file!: ".$this->fname);
    }
  }
  
  function Getfilename(){
    return $this->fname;
  }

  function Version(){
    return "SongBook Compiler 2023.08.15.20:00\n";
  }
  
  /*
   * Help output
   */
  function Help($err =""){
    $nl = $this->cli? PHP_EOL: "<br />";

    $s  = "";
    $s .="syntax: songbook.exe filename <default folder> | -tag <name_of_tag>".$nl;
    $s .= "    filename - file.pro".PHP_EOL;
    $s .= "    <default folder> - path for the file".$nl;
    $s .= "    -tag <name_of_tag> - any other tag for the SongBook file".$nl;
    $s .= "ERROR:".$err;
    if(isset($_SESSION)){
      session_destroy();
    }
    if(isset($argv)){      
      die($s);
    }
  }

  function Test($test= true){
    $this->test = $test;
  }
  /*
   * Read the tags from command line
   */
  function ReadTags(){
    if( isset($argv) ){
      $this->tags_cl[]    = basename($argv[2]);
      $tindex = array_search("-tag",$argv);
      if($tindex){
        for($i = $tindex+1; $i< count($argv); $i++){
          $this->tags_cl[] = $argv[($i)];
        }
      }      
    }else{
      $this->tags_cl[] = basename(getcwd());
    }    
  }
  
 /**
   * Compile the file
   */
  function Compile(){    
    $this->Header();    
    $this->Chords();
    $this->ChordIntoRow();
    $this->LastEmptyRows();
  }
  
  /**
   * Delete last empty rows
   */
  function LastEmptyRows(){
    if(count($this->rows)>0){
      end($this->rows);
      while( empty( trim( current($this->rows) ) ) ) {
        $k = key($this->rows);
        unset($this->rows[$k]);
        end($this->rows);
      }
    }
  }
  /*
   * Load the content os file
   */
  function Load(){
    $this->h = file_get_contents($this->fname);
    if(empty($this->h)) $this->Help("Can not read the file: ".$this->fname);

    $this->size0 = strlen($this->h);
    
    //The rows split into array
    $this->rows = file($this->fname, FILE_IGNORE_NEW_LINES);

    //Delete the first, empty rows
    reset($this->rows);
    while ( strlen ( trim( current($this->rows ) ) ) == 0 ){
      $key = key($this->rows);
      unset( $this->rows[$key] );
      reset($this->rows);
    }
  }
  
  //Make or modify headers
  function Header(){ 
    print("Compile headers\n");
    $this->ReadHeaders();
    $this->MakeTitleSubtitle();
    $this->MakeTags();
  }
  
  function ReadHeaders(){
    $m =[];
    foreach($this->hs AS $i => $head){
      if(preg_match($head, $this->h, $m ) === 1){
        $this->hs[$i] = true;
      }else{
        $this->hs[$i] = false;
      }
    }
  }
  
  /*
   * Make title and subtitle
   */  
  function MakeTitleSubtitle(){ 
    reset($this->rows);
    $k   = key($this->rows);
    $c   = explode("-", $this->rows[$k] ); 
    if(count($c)>1)
    {
      $st = str_replace(['_'],[' '],trim($c[0]));
      $t  = str_replace(['_'],[' '],trim($c[1]));
      unset($this->rows[$k]);     
    }else{
      $t  = "";
      $st = "";
    }

    //Title
    if(!empty($t) && !$this->hs["t"] )
    {  
      $this->headers['t'] = "{t: ".$t."}";  
      foreach($this->rows AS $i =>$row){
        if( preg_match("/^\{t:/i", $row) || preg_match("/^\{title:/i", $row) )
        {
          unset($this->rows[$i]);
        }
      }
    }else{
      $ok = False;
      foreach($this->rows AS $i => $row){
        if(preg_match("/^\{t:.*\}/i",$row,$m,PREG_OFFSET_CAPTURE)){
          $this->headers["t"] = $m[0][0];
          unset($this->rows[$i]);
          $ok = True;
        }
        if(preg_match("/^\{title:.*\}/i",$row,$m,PREG_OFFSET_CAPTURE)){
          $this->headers["t"] = $m[0][0];
          unset($this->rows[$i]);
          $ok = true;
        }        
      }
      if(!$ok){
        $this->headers["t"] = "{t:".basename($this->fname,".pro")."}";
      }
    }
    
    //Subtitle    
    if(!empty($st) && !$this->hs["st"] ){ 
      $this->headers['st'] = "{st: ".$st."}";
      $this->headers["artist"] ="{artist: ".$st."}";
      foreach($this->rows AS $i =>$row){
        if( preg_match("/^\{st:/i", $row) || preg_match("/^\{subtitle:/i", $row) )
        {
          unset($this->rows[$i]);
        }
      }
    }else{
      $ok = False;
      foreach($this->rows AS $i => $row){
        if(preg_match("/^\{st:.*\}/i",$row,$m,PREG_OFFSET_CAPTURE)){
          $this->headers["st"] = $m[0][0];
          unset($this->rows[$i]);
        }
        if(preg_match("/^\{subtitle:.*\}/i",$row,$m,PREG_OFFSET_CAPTURE)){
          $this->headers["st"] = $m[0][0];
          unset($this->rows[$i]);
        }
      }
      if(!$ok){
        $this->headers["st"] = "{st:}";
      }
    }

  }

  /*
   * Make tags
   */
  function MakeTags(){
    $m = array();
    foreach($this->tags_cl AS $tag){
      if(!preg_match("/\{tag:".$tag."}/i", $this->h,$m, PREG_OFFSET_CAPTURE))
      {
        $this->headers[$tag] = "{tag: ".$tag."}";
      }
    }
  }

  /**
   * Write out the new file 
   */
  function Write(){
    if($this->cli){
      $nl = PHP_EOL;     
    }else{
     $nl = "<br/>";
    }
    $heads = "";
    foreach($this->headers AS $h){
      if(!empty($h)){
        $heads .= $h.PHP_EOL;
      }
    }
    $chords = ""; 
    foreach($this->rows AS $row){
      $chords .= $row.PHP_EOL;
    }
    $str   = $heads.$chords;
    $this->size1 = strlen($str);
    $msg ="";
    //Test version
    if($this->test){
      file_put_contents( $this->ChangeFname("_test"),$str);
      $msg = "Test version";
      die($nl.$msg.$nl);
    }
    
    //If the new file larger than original
    if($this->size1 > $this->size0){
      file_put_contents( $this->fname, $str);
      $msg = "OK. The size increased";
    }else{
      rename($this->fname, $this->ChangeFname("_bak"));
      file_put_contents( $this->fname,$str);
      $msg = "OK. Backu + new version";
    } 
    die($nl.$msg.$nl);
  }
  
  function ChangeFname($c="_test"){
    $p = pathinfo($this->fname) ;
    $f = $p['filename'];
    $e = $p['extension'];
    return $f.$c.".".$e;
  }
  
  //Write out the start of tag and the end of tab tab
  function Tabs($i){    
    if(!isset($this->rows[$i])) return true;
    
    if( ( !$this->sot) && (
            strpos($this->rows[$i], "e|")  === 0 || //Sz�l� git�r
            strpos($this->rows[$i], "E|")  === 0 ||
            strpos($this->rows[$i], "D#|") === 0 || 
            strpos($this->rows[$i], "G|")  === 0 || //Bass git�r
            strpos($this->rows[$i], "F#|")  === 0
        ) 
    ){
      $this->rows[$i] = "{sot}".PHP_EOL.$this->rows[$i];
      $this->sot= true;
      return true;
    }    
    if($this->sot && empty(trim($this->rows[$i])) ){
      unset($this->rows[$i]);
      return true;
    }
    //Write out the end of tab
    if( $this->sot && ( 
        strpos($this->rows[$i], "E|") === 0 || 
        strpos($this->rows[$i], "D#|") === 0
        ) 
    ) 
    {
      $this->rows[$i] .= PHP_EOL."{eot}".PHP_EOL;
      $this->sot = false;
      return true;
    }
    return false;
  }
  /**
   * Notices
   * @param unknown $i
   * @return boolean
   */
  function Comment($i){
    $ok = false;
    if(!isset($this->rows[$i])) 
      return true;
    foreach($this->words As $word){
      if(strlen($this->rows[$i]) < strlen($word)) 
        continue;      
      $ok = $ok || !(false === strpos($this->rows[$i],"[".$word."]"));      
      $this->rows[$i] = str_ireplace("[".$word."]", "{c:".$word."}",$this->rows[$i]);
    } 
    
    if(!$ok && ( strpos($this->rows[$i],"{c:") !== false )  )
    {
      $ok = true;
    }
    if($ok){
      $j=1;
      while( isset($this->rows[$i-$j] ) && empty( trim( $this->rows[$i-$j]  ) ) ){
        unset($this->rows[$i-$j]);
        $j++;
      }
      
      $j=1;
      while( isset($this->rows[$i+$j] ) && empty( trim( $this->rows[$i+$j]  ) ) ){
        unset($this->rows[$i+$j]);
        $j++;
      }
      $this->rows[$i] = PHP_EOL.$this->rows[$i];
    }
    return $ok;
  }
  
  /**
   * Chords
   */
  function Chords(){
    $nl = $this->cli ? PHP_EOL : "<br />";

    print("Compile Chords and others...".$nl."\n");
    $this->max = count($this->rows);
    $pc = -1;
    foreach($this->rows AS $i => $row){
      $this->chg = array();
      $percent =  (int) (100 * $i / $this->max);      
      if( ($percent != $pc) && ($percent % 10) == 0){ 
        print($nl.$percent."% ");
        $pc = $percent;
      }else{
        print(".");
      }
      //jump over the tabs
      if(preg_match("/{sot}/i",$row) ){
        $this->sot = true;
        continue;
      }
    
      if(preg_match("/{eot}/i",$row) ){
        $this->rows[$i] ="";
        $this->sot = false;
        continue;
      }
      
      if($this->Tabs($i))
        continue;

      //jump over the comment      
      if($this->Comment($i)) 
        continue;
     
     // loop for the chords in a row      
      $row = $this->pre.$row.$this->post;
      
      $this->CompileChords($row, $i);
      
      $row = str_replace([$this->pre , $this->post], ["",""], $row );
      
      //clear unneeded rows
      if(($this->rows[$i] != $row) && isset($this->rows[$i-1]) && (empty(trim($this->rows[$i-1])) )){
        unset($this->rows[$i-1]);
      }
      if(($this->rows[$i] != $row) && isset($this->rows[$i+1]) && (empty(trim($this->rows[$i+1])) )){
        unset($this->rows[$i+1]);
      }

      $this->rows[$i] = $row;
    } //end of foreach    
  }
  /*
   * Compile chords
   */
  function CompileChords( &$row, $i){
    if(strpos($row,"E  A D E A")>0){
      $debug = true;
    }
 
    foreach ($this->Tones AS $tone)
    {
      foreach($this->chords AS $idx => $chord)
      {
        //A letter at the start of row
        if( $tone == "A"){
          if( preg_match("/A[ \t][jklmnpqrstuvwxzJKLMNPQRSTVWXZ]/",$row) && (strlen($row)>10 ) ){
             continue;
          }
        }
      
        $tc = $tone.$chord;
      
        //chord is begin and end of line
        if ($row == $this->pre.$tc.$this->post) {
          $row = "[".$tc."]";
          continue;
        }

        // Beginning of Chord/Bass $tc = Tone+Chord, pl: Abm/
        $s = [ $tc."/",
               " ".$tc."/" ,
               "\t".$tc."/" , 
               ",".$tc."/" 
             ];
        
        $r = [ "[".$tc."/", 
               " [".$tc."/", 
               "\t[".$tc."/", 
               ",[".$tc."/" 
             ];
        $row = str_replace($s, $r, $row);

        //Chord/Bass   Ab     
        $s = [ "/".$tone." " , 
               "/".$tone."\t" , 
               "/".$tone.",", 
               "/".$tone.$this->post 
            ];
        $r = [ "/".$tone."] ", 
               "/".$tone."]\t", 
               "/".$tone."],", 
               "/".$tone."]".$this->post
             ];
        $row = str_replace($s, $r, $row);

        //chord beginning of line or end of line
        $s = [ $this->pre.$tc." " , 
               $this->pre.$tc."\t" , 
               " ".$tc.$this->post, 
               "\t".$tc.$this->post
            ];
        $r = [ $this->pre."[".$tc."] ", 
               $this->pre."[".$tc."]\t", 
               " [".$tc."]".$this->post, 
               "\t[".$tc."]".$this->post
            ];
        $row = str_replace($s, $r ,$row);

        //Chord alone \	t, " "
        $s = [ 
               "(".$tc.")",
               " ".$tc." ",
               " ".$tc.",", 
               " ".$tc.":",
               " ".$tc."|", 
               " ".$tc."\t",
               "|".$tc." ",
               "|".$tc."|", 
               " ".$tc."-" ,
               "\t".$tc."\t",
               "\t".$tc." ",
               "\t".$tc.",",
               "\t".$tc.":"
        ];
        $r = [ 
               "[(".$tc.")]",
               " [".$tc."] ", 
               " [".$tc."],", 
               " [".$tc."]:",
               " [".$tc."]|",
               " [".$tc."]\t",
               "|[".$tc."] ",
               "|[".$tc."]|",
               " [".$tc."]-",
               "\t[".$tc."]\t",
               "\t[".$tc."] ",
               "\t[".$tc."],",
               "\t[".$tc."]:"
            ];
        $row = str_replace($s, $r ,$row);

        $row = preg_replace("/(^|\W|[^áéíóöőúű])".$tc." /", "[".$tc."] " , $row);
        $row = preg_replace("/(^|\W|[^áéíóöőúű])".$tc."\t/","[".$tc."]\t" , $row);
      }
    }
  	$row = str_replace(array("[[","]]"),array("[","]"), $row);
  }
  
  function ChordIntoRow(){
    $m     = array();
    $place = 0;
    $rowp ="";    
    $ii =0;
    foreach( $this->rows AS $i => $row ){
      
      if( empty( trim($row) ) ||
          strpos($row,"{c:"  ) !== FALSE ||
          strpos($row,"{sot}") !== FALSE ||
          strpos($row,"{eot}") !== FALSE ||
          strpos($row,"e|"   ) !== FALSE ||
          strpos($row,"b|"   ) !== FALSE ||
          strpos($row,"G|"   ) !== FALSE ||
          strpos($row,"D|"   ) !== FALSE ||
          strpos($row,"A|"   ) !== FALSE ||
          strpos($row,"E|"   ) !== FALSE )
      {   
        $rowp = $row;
        continue;
      }
      //there are chords in the current row
      $ok1 = preg_match("/\[[\w\#\/]*\][\W]/m",$row." ");
      if($ok1 != 0){
        $rowp = $row;
        $ii = $i;
        continue;
      }
      //is there some records in the previous row some?
      $ok2 = preg_match_all("/\[[\w\#\/]*\][\W]/m",$rowp." ",$m, PREG_OFFSET_CAPTURE);
      if($ok2 == 0){
        $rowp = $row;
        $ii = $i;
        continue;
      }
      
      $signs = $m[0];
      /* 
      $d = 2*count($signs)-2; //az akkordok teljes hossza a [] jelekkel együtt
      $signs = array_reverse($signs);
      foreach($signs AS $si => $sign){
        $chord = trim($sign[0]);
        $place = $sign[1]-$d;
        $row   = substr($row,0,$place).$chord.substr($row, $place);
        $d-=2; 
      }     
      */
      $row = str_pad($row, 100);
      for( $j = count($signs)-1; $j >=0 ;$j--){
        $chord = trim($signs[$j][0]);
        $place = $signs[$j][1]-$j*2;
        $row   = substr($row,0,$place).$chord.substr($row, $place);
      }
      $row = rtrim($row);
      
      $this->rows[$i] = $row;
      unset($this->rows[$ii]);
      $ii = $i;
    }    
  }
}